@@ -1,18 +1,23 @@
 # SPDX-License-Identifier: MIT
 
+import contextlib
 import email
+import functools
 import glob
 import gzip
+import importlib.util
 import io
+import itertools
 import os
 import os.path
+import shutil
 import subprocess
 import sys
 import tarfile
 import typing
 import warnings
 
-from typing import IO, Dict, List, Mapping, Optional, Sequence, Tuple, Union
+from typing import IO, ContextManager, Dict, Iterable, Iterator, List, Mapping, Optional, Sequence, Set, Tuple, Union
 
 import packaging.markers
 import packaging.requirements
@@ -20,6 +25,7 @@ import packaging.version
 import toml
 
 import trampolim._metadata
+import trampolim._tasks
 
 
 if sys.version_info < (3, 8):
@@ -50,6 +56,45 @@ class TrampolimWarning(Warning):
     '''Backend warning.'''
 
 
+def load_file_module(name: str, path: str) -> object:
+    spec = importlib.util.spec_from_file_location(name, path)
+    if not spec.loader:  # pragma: no cover
+        raise ImportError(f'Unable to import `{path}`: no loader')
+    module = importlib.util.module_from_spec(spec)
+    spec.loader.exec_module(module)  # type: ignore
+    return module
+
+
+@contextlib.contextmanager
+def cd(path: str) -> Iterator[None]:
+    cwd = os.getcwd()
+    os.chdir(path)
+
+    try:
+        yield
+    finally:
+        os.chdir(cwd)
+
+
+def ensure_empty_dir(path: str) -> None:
+    shutil.rmtree(path, ignore_errors=True)
+    os.makedirs(path)
+
+
+def copy_to_dir(files: Iterable[str], out: str) -> None:
+    for file in files:
+        dirpath = os.path.dirname(file)
+        if dirpath:
+            destdirpath = os.path.join(out, dirpath)
+            os.makedirs(destdirpath, exist_ok=True)
+            shutil.copystat(dirpath, destdirpath)
+        shutil.copy2(
+            file,
+            os.path.join(out, file),
+            follow_symlinks=False,
+        )
+
+
 class Project():
     _VALID_DYNAMIC = [
         'version',
@@ -71,6 +116,9 @@ class Project():
 
         self.version  # calculate version
 
+        self._path = os.getcwd()
+        self._extra_binary_source: Set[str] = set()
+
         # warn users about test/tests modules -- they probably don't want them installed!
         for module in ('test', 'tests'):
             if module in self.root_modules:
@@ -79,21 +127,81 @@ class Project():
                     TrampolimWarning,
                 )
 
-    @property
-    def source(self) -> List[str]:
-        '''Project source.'''
-        source = []
+        # copy distribution source to working directory
+        self._dist_srcpath: str = os.path.abspath(os.path.join('.trampolim', 'dist-source'))
+        ensure_empty_dir(self._dist_srcpath)
+        copy_to_dir(
+            itertools.chain(self.distribution_source, self.build_system_source),
+            self._dist_srcpath,
+        )
+        # copy binary source to working directory
+        self._bin_srcpath: str = os.path.abspath(os.path.join('.trampolim', 'bin-source'))
+        ensure_empty_dir(self._bin_srcpath)
+        copy_to_dir(
+            itertools.chain(self.binary_source, self.build_system_source),
+            self._bin_srcpath,
+        )
 
-        # TODO: ignore files not escaped as specified in PEP 427
-        for module in self.root_modules:
-            if module.endswith('.py'):  # file
-                source.append(module)
-            else:  # dir
-                source += [
-                    path for path in glob.glob(os.path.join(module, '**'), recursive=True)
-                    if os.path.isfile(path) and not path.endswith('.pyc')
-                ]
-        return source
+        # collect tasks
+        if os.path.isfile('.trampolim.py'):
+            config = load_file_module('trampolim_config', '.trampolim.py')
+            self._tasks = [
+                getattr(config, attr)
+                for attr in dir(config)
+                if not attr.startswith('_') and isinstance(
+                    getattr(config, attr), trampolim._tasks.Task
+                )
+            ]
+        else:
+            self._tasks = []
+
+    @functools.lru_cache(maxsize=None)
+    def run_tasks(self) -> None:
+        '''Runs the project build tasks.
+
+        If the ``SOURCE_DATE_EPOCH`` environment variable is present, extra
+        sources files added by tasks will have their atime and mtime set to it.
+        '''
+        with self.cd_binary_source():
+            for task in self._tasks:
+                print(f'> Running `{task.name}`')
+                session = trampolim._tasks.Session(self)
+                task.run(session)
+                self._extra_binary_source |= set(session.extra_source)
+                # TODO: print summary
+
+        # set the extra source atime and mtime
+        source_date_epoch = os.environ.get('SOURCE_DATE_EPOCH')
+        if source_date_epoch:
+            timestamp = int(source_date_epoch)
+            with self.cd_binary_source():
+                for file in self._extra_binary_source:
+                    os.utime(file, times=(timestamp, timestamp))
+
+    def cd_dist_source(self) -> ContextManager[None]:
+        return cd(self._dist_srcpath)
+
+    def cd_binary_source(self) -> ContextManager[None]:
+        return cd(self._bin_srcpath)
+
+    @property
+    def build_system_source(self) -> Iterable[str]:
+        return iter(filter(None, [
+            'pyproject.toml',
+            '.trampolim.py' if os.path.isfile('.trampolim.py') else None,
+            self.license_file,
+            self.readme_file,
+        ]))
+
+    @property
+    def distribution_source(self) -> Iterable[str]:
+        '''Project source -- for source distributions.'''
+        return self.modules_source | self.config_source_include
+
+    @property
+    def binary_source(self) -> Iterable[str]:
+        '''Python package source -- for binary distributions.'''
+        return self.modules_source | self.config_source_include | self._extra_binary_source
 
     @property
     def root_modules(self) -> Sequence[str]:
@@ -113,6 +221,29 @@ class Project():
             return [name]
         raise TrampolimError(f'Could not find the top-level module(s) (looking for `{name}`)')
 
+    @property
+    def modules_source(self) -> Set[str]:
+        '''Full source for the modules in root_modules.'''
+        source = set()
+        # TODO: ignore files not escaped as specified in PEP 427
+        for module in self.root_modules:
+            if module.endswith('.py'):  # file
+                source.add(module)
+            else:  # dir
+                source |= {
+                    path for path in glob.glob(os.path.join(module, '**'), recursive=True)
+                    if os.path.isfile(path) and not path.endswith('.pyc')
+                }
+        return source
+
+    @property
+    def config_source_include(self) -> Set[str]:
+        '''Extra source include paths specified in the config.'''
+        return {
+            os.path.sep.join(path.split('/'))
+            for path in self._trampolim_meta.source_include
+        }
+
     @property
     def name(self) -> str:
         '''Project name.'''
@@ -361,26 +492,33 @@ class SdistBuilder():
             format=tarfile.PAX_FORMAT,  # changed in 3.8 to GNU
         )
 
-        # add pyproject.toml
-        tar.add('pyproject.toml', f'{self.name}/pyproject.toml')
+        with self._project.cd_dist_source():
+            # add pyproject.toml
+            tar.add('pyproject.toml', f'{self.name}/pyproject.toml')
 
-        # add source
-        for source_path in self._project.source:
-            tar.add(source_path, f'{self.name}/{source_path}')
+            # add .trampolim.py
+            try:
+                tar.add('.trampolim.py', f'{self.name}/.trampolim.py')
+            except FileNotFoundError:  # pragma: no cover
+                pass
 
-        # add license
-        if self._project.license_file:
-            tar.add(self._project.license_file, f'{self.name}/{self._project.license_file}')
-        elif self._project.license_text:
-            license_raw = self._project.license_text.encode()
-            info = tarfile.TarInfo(f'{self.name}/LICENSE')
-            info.size = len(license_raw)
-            with io.BytesIO(license_raw) as data:
-                tar.addfile(info, data)
+            # add source
+            for source_path in self._project.distribution_source:
+                tar.add(source_path, f'{self.name}/{source_path}')
 
-        # add readme
-        if self._project.readme_file:
-            tar.add(self._project.readme_file, f'{self.name}/{self._project.readme_file}')
+            # add license
+            if self._project.license_file:
+                tar.add(self._project.license_file, f'{self.name}/{self._project.license_file}')
+            elif self._project.license_text:
+                license_raw = self._project.license_text.encode()
+                info = tarfile.TarInfo(f'{self.name}/LICENSE')
+                info.size = len(license_raw)
+                with io.BytesIO(license_raw) as data:
+                    tar.addfile(info, data)
+
+            # add readme
+            if self._project.readme_file:
+                tar.add(self._project.readme_file, f'{self.name}/{self._project.readme_file}')
 
         # PKG-INFO
         pkginfo = self._project.metadata.as_bytes()
